;; $Id: rucksack.lisp,v 1.27 2009-05-27 14:26:25 alemmens Exp $

#+xcvb
(module
 (:depends-on
  ("package")))

(in-package :rucksack)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Rucksacks: API
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; open-rucksack [Function]
;; close-rucksack [Function]
;; with-rucksack [Macro]
;; current-rucksack [Function]

;; commit [Function]
;; rollback [Function]

(defgeneric add-rucksack-root (object rucksack)
  (:documentation
 "Adds an object to the root set of a rucksack."))

(defgeneric delete-rucksack-root (object rucksack)
  (:documentation
 "Delete an object from the root set of a rucksack."))

(defgeneric map-rucksack-roots (function rucksack)
  (:documentation
 "Applies a function to all objects in the root set of a rucksack."))

(defgeneric rucksack-roots (rucksack)
  (:documentation
 "Returns a list with all objects in the root set of a rucksack.  You
shouldn't modify this list."))

(defgeneric rucksack-root-p (object rucksack)
  (:documentation
   "Returns true iff OBJECT is a member of the root set of a rucksack."))

(defgeneric rucksack-cache (rucksack)
  (:documentation "Returns the cache for a rucksack."))

(defgeneric rucksack-directory (rucksack)
  (:documentation
 "Returns a pathname for the directory that contains all files of a
rucksack."))

(defgeneric rucksack-commit (rucksack)
  (:documentation
 "Ensures that all in-memory data is saved to disk."))

(defgeneric rucksack-rollback (rucksack)
  ;; DO: What does rollback mean exactly here?
  (:documentation "...."))

;;
;;  Class and slot indexing
;;

;; add-class-index (class-designator &key errorp)  [Function]
;; add-slot-index (class-designator slot index-spec &key errorp) [Function]
;; remove-class-index (class-designator &key errorp) [Function]
;; remove-slot-index (class-designator slot &key errorp) [Function]
;; map-class-indexes (function) [Function]
;; map-slot-indexes (function &key class include-subclasses) [Function]


(defgeneric rucksack-update-class-index (rucksack class)
  (:documentation
   "Compares the current class index for CLASS to the class index
that's specified in the :INDEX class option of CLASS.  An obsolete
class index (i.e. a class index that's specified anymore in the class
option) is removed, new class indexes are added."))

(defgeneric rucksack-update-slot-indexes (rucksack class)
  (:documentation
   "Compares the current slot indexes for CLASS to the slot indexes
that are specified in the slot options for the direct slots of CLASS.
Obsolete slot indexes (i.e. slot indexes that are not specified
anymore in the slot options or indexes for slots that don't exist
anymore) are removed, new slot indexes are added."))

(defgeneric rucksack-add-class-index (rucksack class-designator &key errorp))

(defgeneric rucksack-remove-class-index (rucksack class-designator
                                                  &key errorp))

(defgeneric rucksack-class-index (rucksack class-designator &key errorp)
  (:documentation "Returns the class index for a class designator."))

(defgeneric rucksack-map-class-indexes (rucksack function)
  (:documentation
   "FUNCTION must take two arguments: a class name and a class index.
It is called for all class indexes in the specified rucksack."))

(defgeneric rucksack-make-class-index (rucksack class &key index-spec)
  (:documentation
   "Creates a new class index and returns that index.  INDEX-SPEC
specifies the kind of index that must be created (if not supplied, the
rucksack's default class index spec will be used."))


(defgeneric rucksack-add-slot-index (rucksack class-designator slot index-spec
                                     unique-p &key errorp)
  (:documentation
  "Creates a new slot index for the slot designated by
CLASS-DESIGNATOR and SLOT.  The type of index is specified by
INDEX-SPEC.  Returns the new index.  Signals an error if ERRORP is T
and there already is an index for the designated slot."))

(defgeneric rucksack-remove-slot-index (rucksack class-designator slot
                                        &key errorp))



(defgeneric rucksack-slot-index (rucksack class-designator slot
                                 &key errorp include-superclasses)
  (:documentation
 "Returns the slot index for the slot specified by CLASS-DESIGNATOR
and SLOT."))


(defgeneric rucksack-map-slot-indexes (rucksack function
                                       &key class include-subclasses)
  (:documentation
   "FUNCTION must take three arguments: a class name, a slot name and
a slot index.  It is called for all slot indexes in the specified
rucksack.
  CLASS defaults to T, meaning all classes.
  INCLUDE-SUBCLASSES defaults to T."))

(defgeneric rucksack-maybe-index-changed-slot (rucksack
                                               class object slot
                                               old-value new-value
                                               old-boundp new-boundp)
  (:documentation
 "This function is called after a slot has changed.  OLD-VALUE is the
slot's value before the change, NEW-VALUE is the current value.
OLD-BOUNDP is true iff the slot was bound before the change,
NEW-BOUNDP is true iff the slot is currently bound."))

(defgeneric rucksack-maybe-index-new-object (rucksack class-designator object)
  (:documentation
 "Adds the object id of OBJECT to the class index for the class
designated by CLASS-DESIGNATOR.  If there is no such class index, it
does nothing."))

(defgeneric rucksack-map-class (rucksack class function
                                &key id-only include-subclasses)
  (:documentation
 "  FUNCTION is a unary function that gets called for all instances of
the specified class.  Unindexed classes (i.e. classes for which the
:indexed class option is nil) will be skipped.
  If ID-ONLY is T (default is NIL), the function will be called with
object ids instead of 'real' objects.  This can be handy if you want to
do more filtering before actually loading objects from disk.
  INCLUDE-SUBCLASSES defaults to T."))

(defmacro rucksack-do-class ((instance-var class
                              &key
                              (rucksack '*rucksack*)
                              id-only
                              (include-subclasses t))
                             &body body)
  "Evaluate BODY for each instance of CLASS, with INSTANCE-VAR
successively bound to each instance.  See the documentation of
RUCKSACK-MAP-CLASS for more details."
  (check-type instance-var symbol)
  `(rucksack-map-class ,rucksack ,class
                       (lambda (,instance-var) ,@body)
                       :id-only ,id-only
                       :include-subclasses ,include-subclasses))


(defgeneric rucksack-map-slot (rucksack class slot function
                              &key equal min max include-min include-max order
                              include-subclasses)
  (:documentation
 " FUNCTION is a unary function that gets called for all instances of
the specified class that have a slot value matching the EQUAL, MIN,
MAX INCLUDE-MIN and INCLUDE-MAX arguments (see the documentation of
MAP-INDEX for a description of these arguments).
  ORDER can be either :ASCENDING (default) or :DESCENDING; currently,
the specified order will be respected for instances of one class but
not across subclasses.
  If ID-ONLY is T (default is NIL), the function will be called with
object ids instead of 'real' objects.  This can be handy if you want to
do more filtering before actually loading objects from disk.
  INCLUDE-SUBCLASSES defaults to T."))

(defmacro rucksack-do-slot ((instance-var class slot
                             &rest args
                             &key (rucksack '*rucksack*)
                             equal min max include-min include-max
                             order include-subclasses)
                            &body body)
  "Evaluate BODY for each instance of CLASS where SLOT has the
specified value. INSTANCE-VAR will be bound successively to each
instance.  See the documentation of RUCKSACK-MAP-SLOT for more
details."
  (declare (ignorable equal min max include-min include-max order
                      include-subclasses))
  (check-type instance-var symbol)
  `(rucksack-map-slot ,rucksack ,class ,slot
                      (lambda (,instance-var) ,@body)
                      ,@(sans args ':rucksack)))


#+later
(defgeneric rucksack-map-objects (rucksack class-designator function
                                           slots filter order)
  (:documentation
 " Applies FUNCTION to all instances of the class designated by
CLASS-DESIGNATOR for which the criteria specified by SLOTS and
CRITERIA hold.
  SLOTS is a list of slot names.  FILTER is a filter expression that can
refer to the slot names.
  Example of a filter expression: (and (= age 20) (string= city \"Hamburg\"))
"))


(defgeneric rucksack-delete-object (rucksack object)
  (:documentation
   "Removes OBJECT from RUCKSACK, i.e. removes object from the
rucksack roots (if it is a root) and from all class and slot indexes
in which it appears."))


(defgeneric object-id (object)
  (:documentation "Returns the object id of a persistent-object or
persistent-data."))

;;; The following functions need to be documented.
(defgeneric class-index (class))
(defgeneric saved-slots (object))
(defgeneric deserialize-contents (marker stream))
(defgeneric close-heap (heap))
(defgeneric finish-heap-output (heap))
(defgeneric heap-size (heap))
(defgeneric expand-heap (heap block-size))
(defgeneric reset-allocation-counter (heap))
(defgeneric size-class (size heap))
(defgeneric size-class-block-size (size-class heap))
(defgeneric free-list-start (heap &optional (size-class)))
(defgeneric (setf free-list-start) (pointer heap &optional size-class))
(defgeneric free-list-empty-p (size-class heap))
(defgeneric block-header-size (heap))
(defgeneric block-header (block heap))
(defgeneric (setf block-header) (value block heap))
(defgeneric (setf block-size) (size block heap))
(defgeneric deallocate-block (block heap))
(defgeneric expand-free-list (size-class heap))
(defgeneric find-block (min-size heap))
(defgeneric carve-up-block-for-free-list (size-class block size heap))
(defgeneric heap-info (heap))
(defgeneric free-list-info (size-class heap))
(defgeneric load-heap-end (heap))
(defgeneric buffer-count (buffer))
(defgeneric reset-buffer (buffer))
(defgeneric save-buffer (buffer stream &key))
(defgeneric load-buffer (buffer stream nr-octets &key))
(defgeneric nr-persistent-slots (schema))
(defgeneric fresh-schema-id (table))
(defgeneric find-schema-for-id (table id &key))
(defgeneric find-schema-for-class (table class))
(defgeneric old-schemas-for-class (table class))
(defgeneric find-or-create-schema-for-object (table object))
(defgeneric create-schema (table class version &optional (persistent-slots)))
(defgeneric compute-persistent-slot-names (class object))
(defgeneric add-schema (table schema))
(defgeneric save-schema-table (table))
(defgeneric save-schema-table-if-necessary (table))
(defgeneric maybe-update-schemas (table class))
(defgeneric gc-work-for-size (heap size))
(defgeneric free-space (heap))
(defgeneric handle-written-object (object-id block heap))
(defgeneric work-left (heap))
(defgeneric work-done (heap))
(defgeneric max-work (heap))
(defgeneric nr-object-bytes (heap))
(defgeneric nr-heap-bytes (heap))
(defgeneric collect-garbage (heap))
(defgeneric finish-garbage-collection (heap))
(defgeneric collect-some-garbage (heap amount))
(defgeneric mark-some-objects-in-table (heap amount))
(defgeneric mark-some-roots (heap amount))
(defgeneric mark-root (heap object-id))
(defgeneric load-block (heap block &key buffer skip-header))
(defgeneric sweep-some-heap-blocks (heap amount))
(defgeneric block-alive-p (object-table object-id block))
(defgeneric sweep-some-object-blocks (heap amount))
(defgeneric cache-full-p (cache))
(defgeneric cache-room (cache))
(defgeneric cache-touch-object (object cache))
(defgeneric make-room-in-cache (cache))
(defgeneric add-to-queue (object-id cache))
(defgeneric maybe-dereference-proxy (proxy))
(defgeneric persistent-data-read (function data &rest args))
(defgeneric persistent-data-write (function data value &rest args))
(defgeneric p-aref (array &rest indices))
(defgeneric (setf p-aref) (new-value array &rest indices))
(defgeneric p-array-dimensions (array))
(defgeneric p-car (cons))
(defgeneric (setf p-car) (value cons))
(defgeneric p-cdr (cons))
(defgeneric (setf p-cdr) (value cons))
(defgeneric p-consp (object))
(defgeneric p-endp (object))
(defgeneric p-length (vector))
(defgeneric p-find (value vector &key))
(defgeneric p-position (value vector &key))
(defgeneric p-replace (vector-1 vector-2 &key))
(defgeneric p-delete-if (test list &key))
(defgeneric p-nreverse (object))
(defgeneric scan-object (object-id block gc))
(defgeneric save-object (object object-id cache transaction-id previous-block &key))
(defgeneric load-object (object-id transaction cache))
(defgeneric btree-key< (btree))
(defgeneric btree-key= (btree))
(defgeneric btree-key>= (btree))
(defgeneric btree-key<= (btree))
(defgeneric btree-key> (btree))
(defgeneric btree-value= (btree))
(defgeneric bnode-nr-keys (node))
(defgeneric node-max-depth (node))
(defgeneric node-min-depth (node))
(defgeneric btree-depths (btree))
(defgeneric btree-balanced-p (btree))
(defgeneric index-similar-p (index-1 index-2))
(defgeneric map-index-data (index function &rest args &key))
(defgeneric index-data-insert (index key value &key if-exists))
(defgeneric index-data-delete (index key value &key if-does-not-exist))
(defgeneric rucksack-roots-pathname (rucksack))
(defgeneric class-index-table (rucksack))
(defgeneric slot-index-tables (rucksack))
(defgeneric transaction-nr-dirty-objects (transaction))
(defgeneric transaction-touch-object (transaction object object-id))
(defgeneric finish-all-output (rucksack))
(defgeneric save-dirty-object (object cache transaction object-id &key))
